@@ -167,7 +167,6 @@ static xmlNodePtr _domNewItem(xmlNodePtr item) {
 static xmlNodePtr _domReferenceItem(xmlNodePtr item) {
     xmlNodePtr owner = _domItemOwner(item);
     if (owner != NULL) {
-        assert(owner->type != XML_NAMESPACE_DECL);
         xml6_node_add_reference(owner);
     }
     return item;
@@ -176,7 +175,7 @@ static xmlNodePtr _domReferenceItem(xmlNodePtr item) {
 static xmlNodePtr _domUnreferenceItem(xmlNodePtr item) {
     xmlNodePtr owner = _domItemOwner(item);
     if (owner != NULL) {
-        xml6_node_remove_reference(item);
+        xml6_node_remove_reference(owner);
     }
     return owner;
 }
@@ -490,16 +489,15 @@ domXPathCtxtSetNode(xmlXPathContextPtr ctxt, xmlNodePtr node) {
     if (node != oldNode) {
         xmlDocPtr doc = node ? node->doc : NULL;
         if (ctxt->doc != doc) {
-            if (ctxt->doc != NULL)
+            if (ctxt->doc != NULL) {
                 XML6_FAIL(node, "changing XPath Context documents is not supported");
-
-            ctxt->doc = doc;
-
-            if (doc) {
-                xmlNsPtr *ns = _domXPathCtxtRegisterNS(ctxt, node);
-                if (ns != NULL)
-                     xmlFree(ns);
             }
+            ctxt->doc = doc;
+        }
+
+        xmlNsPtr *ns = _domXPathCtxtRegisterNS(ctxt, node);
+        if (ns != NULL) {
+            xmlFree(ns);
         }
         ctxt->node = node;
     }
